
interface BoolClosure {
  boolean m(Object o);
}
interface BinaryClosure {
  Object m(Object o1, Object o2);
}
class Unimplemented extends Error {}

class MLList {
  final Object hd;
  final MLList tl;
  MLList(Object _hd, MLList _tl) {
    hd = _hd;
    tl = _tl;
  }
  static MLList cons(Object hd, MLList tl) {
    return new MLList(hd,tl);
  }
  static int length(MLList lst) {
    if(lst == null)
      return 0;
    return 1+length(lst.tl);
  }
  static MLList append(MLList lst1, MLList lst2) {
    if(lst1 == null)
      return lst2;
    return cons(lst1.hd, append(lst1.tl, lst2));
  }
  static MLList filter(MLList lst, BoolClosure f) {
    if(lst == null)
      return null;
    if(f.m(lst.hd))
      return cons(lst.hd, filter(lst.tl,f));
    return filter(lst.tl,f);
  }
  static Object fold_left(MLList lst, BinaryClosure f, Object v) {
    if(lst == null)
      return v;
    return fold_left(lst.tl, f, f.m(v,lst.hd));
  }
}

class NonNulls implements BoolClosure {
  public boolean m(Object o) { return o != null; }
}
class NumsLessThan implements BoolClosure {
  int x;
  NumsLessThan(int _x) { x = _x; }
  public boolean m(Object o) { 
    return o != null && o instanceof Integer && ((Integer)o).intValue() < x; 
  }
}
class NumNotNull implements BinaryClosure {
  public Object m(Object o1, Object o2) {
    if(o1 == null) return o2;
    return new Integer(((Integer)o2).intValue() + 1);
  }
}
class SumNumsLessThan implements BinaryClosure {
  int x;
  SumNumsLessThan(int _x) { x = _x; }
  public Object m(Object o1, Object o2) {
    if(o1 == null || !(o1 instanceof Integer) || ((Integer)o1).intValue() >= x)
      return o2;
    return new Integer(((Integer)o2).intValue() + ((Integer)o1).intValue());
  }
} 
class C {
  static MLList noNulls(MLList lst) {
    return MLList.filter(lst,new NonNulls());
  }
  static MLList numsLessThan(MLList lst, int x) {
    return MLList.filter(lst, new NumsLessThan(x));
  }
  static int numNotNull(MLList lst) {
    Object ans = MLList.fold_left(lst, new NumNotNull(), new Integer(0));
    return ((Integer)ans).intValue();
  }
  static int sumNumsLessThan(MLList lst, int x) {
    Object ans = MLList.fold_left(lst, new SumNumsLessThan(x), new Integer(0));
    return ((Integer)ans).intValue();
  }
}
