(* extra credit maybe typed or maybe multimethods or ... *)

type field = string * (obj ref)
and  meth  = string * (obj -> obj list -> obj)
and  obj   = O of (field list) * (meth list)
           | N of float

exception Unimplemented
exception ObjectExpected
exception IntExpected
exception FieldNotFound of string * obj
exception MethodNotFound of string * obj

let newObj () : obj = O([],[])

let fields (obj : obj) =
  match obj with
    O(fs,_) -> fs
  | N _ -> raise ObjectExpected
let methods (obj : obj) =
  match obj with
    O(_,ms) -> ms
  | N _ -> raise ObjectExpected
let num (obj : obj) =
  match obj with
    N n -> n
  | O _ -> raise IntExpected

let getSetHelp (s : string) (obj : obj) (f : obj ref -> 'a) =
  let rec loop lst =
    match lst with
      [] -> raise (FieldNotFound(s,obj))
    | (s2,r)::tl -> if s=s2 then f r else loop tl in
  loop (fields obj)

let get (s : string) (obj : obj) : obj =
  getSetHelp s obj (!)

let set (s : string) (obj : obj) (v : obj) : unit =
  getSetHelp s obj (fun r -> r := v)

let findAndSend (s : string) (rcvr : obj) (ms : meth list) (args : obj list) =
  let rec loop lst =
    match lst with
      [] -> raise (MethodNotFound(s,rcvr))
    | (s2,m)::tl -> if (s=s2) then m rcvr args else loop tl in
  loop ms

let send (s : string) (rcvr : obj) (args : obj list) =
  findAndSend s rcvr (methods rcvr) args

let getter s this args = get s this
let setter s this args = (set s this (List.hd args); this)
  
let newPoint (x : obj) (y : obj) : obj =
  let super = newObj () in
  let fs = ("x",ref x)::("y",ref y)::(fields super)in
  let distToOrigin1 this args =
    let n1 = num (get "x" this) in
    let n2 = num (get "y" this) in
    N (sqrt ((n1 *. n1) +. (n2 *. n2))) in
  let distToOrigin2 this args =
    let n1 = num (send "getX" this []) in
    let n2 = num (send "getY" this []) in
    N (sqrt ((n1 *. n1) +. (n2 *. n2))) in
  let ms = ("getX",getter "x")
           ::("getY",getter "y")
           ::("setX",setter "x")
           ::("setY",setter "y")
           ::("distToOrigin1",distToOrigin1)
           ::("distToOrigin2",distToOrigin2)
           ::(methods super) in
  O(fs,ms)

let newPoint3D (x : obj) (y : obj) (z : obj) : obj =
  let super = newPoint x y in
  let fs = ("z",ref z)::(fields super) in
  let distToOrigin1 this args =
    let n1 = num (get "x" this) in
    let n2 = num (get "y" this) in
    let n3 = num (get "z" this) in
    N (sqrt ((n1 *. n1) +. (n2 *. n2) +. (n3 *.n3))) in
  let distToOrigin2 this args =
    let n1 = num (send "getX" this []) in
    let n2 = num (send "getY" this []) in
    let n3 = num (send "getZ" this []) in
    N (sqrt ((n1 *. n1) +. (n2 *. n2) +. (n3 *.n3))) in
  let distToOrigin3 this args =
    let n  = num (findAndSend "distToOrigin1" this (methods super) []) in
    let n3 = num (get "z" this) in
    N (sqrt ((n *. n) +. (n3 *. n3))) in
  let ms = ("getZ",getter "z")
           ::("setZ",setter "z")
           ::("distToOrigin1",distToOrigin1)
           ::("distToOrigin2",distToOrigin2)
           ::("distToOrigin3",distToOrigin3)
           ::(methods super) in
  O(fs,ms)

let toPolar x y = (* return (r,theta) *)
  (sqrt ((x *. x) +. (y *. y)),
   atan (y /. x))
let toCartesian r theta = (* return (x,y) *)
  (r *. (cos theta),
   r *. (sin theta))

let newPolarPoint (x : obj) (y : obj) : obj =
  let super = newPoint x y in
  let r,theta = toPolar (num x) (num y) in
  let fs = ("r",ref (N r))::("theta",ref (N theta))::(fields super) in
  let getX this args = 
    let x,y  = toCartesian (num(get "r" this)) (num(get "theta" this)) in N x in
  let getY this args = 
    let x,y  = toCartesian (num(get "r" this)) (num(get "theta" this)) in N y in
  let setX this args = 
    let x,y  = toCartesian (num(get "r" this)) (num(get "theta" this)) in
    let r,th = toPolar (num (List.hd args)) y in
    (set "r" this (N r)); (set "theta" this (N th)); this in
  let setY this args = 
    let x,y  = toCartesian (num(get "r" this)) (num (get "theta" this)) in
    let r,th = toPolar x (num (List.hd args)) in
    (set "r" this (N r)); (set "theta" this (N th)); this in
  let ms = ("getX",getX)
           ::("getY",getY)
           ::("setX",setX)
           ::("setY",setY)
           ::(methods super) in 
  O(fs,ms)

let newPolarPointB (x : obj) (y : obj) : obj =
  let super = newPolarPoint x y in
  O(fields super, ("distToOrigin1",getter "r")::(methods super))

(* testing *)
let printNum obj = print_string (string_of_float (num obj)); print_newline()

let p1 = newPoint (N 3.0) (N 4.0)
let _ = printNum (send "distToOrigin1" p1 [])
let _ = printNum (send "distToOrigin2" p1 [])
let _ = (send "setY" p1 [N 3.0])
let _ = printNum (send "distToOrigin1" p1 [])
let _ = printNum (send "distToOrigin2" p1 [])

let p2 = newPoint3D (N 0.0) (N 4.0) (N 3.0)
let _ = printNum (send "distToOrigin1" p2 [])
let _ = printNum (send "distToOrigin2" p2 [])
let _ = printNum (send "distToOrigin3" p2 [])
let _ = (send "setY" p2 [N 3.0])
let _ = printNum (send "distToOrigin1" p2 [])
let _ = printNum (send "distToOrigin2" p2 [])
let _ = printNum (send "distToOrigin3" p2 [])

let p3 = newPolarPoint (N 3.0) (N 4.0)
let _ = printNum (send "distToOrigin1" p3 [])
let _ = printNum (send "distToOrigin2" p3 [])
let _ = (send "setY" p3 [N 3.0])
let _ = printNum (send "distToOrigin1" p3 [])
let _ = printNum (send "distToOrigin2" p3 [])

let p4 = newPolarPointB (N 3.0) (N 4.0)
let _ = printNum (send "distToOrigin1" p4 [])
let _ = printNum (send "distToOrigin2" p4 [])
let _ = (send "setY" p4 [N 3.0])
let _ = printNum (send "distToOrigin1" p4 [])
let _ = printNum (send "distToOrigin2" p4 [])
